from functools import reduce, partial
from sismic.exceptions import StatechartError

import abc
import pyparsing  # type: ignore
from typing import Callable, Optional, List


class Sequence(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def evaluate(self, force=False) -> Optional[bool]:
        raise NotImplementedError()

    def __repr__(self):
        return '%s()' % self.__class__.__name__


class SequenceItem(Sequence):
    def __eq__(self, other):
        return self.__class__ == other.__class__


class SequenceUnaryOperator(Sequence):
    def __init__(self, sequence: Sequence) -> None:
        self._sequence = sequence

    @property
    def sequence(self):
        return self._sequence

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.sequence)

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.sequence == other.sequence


class SequenceBinaryOperator(Sequence):
    def __init__(self, left_sequence: Sequence, right_sequence: Sequence) -> None:
        self._left = left_sequence
        self._right = right_sequence

    @property
    def left(self):
        return self._left

    @property
    def right(self):
        return self._right

    def __repr__(self):
        return '%s(%s, %s)' % (self.__class__.__name__, self.left, self.right)

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.left == other.left and self.right == other.right


# ################################################################


class SequenceCondition(SequenceItem):
    def __init__(self, code: str, eval_func: Callable[[str], bool]=None) -> None:
        super().__init__()
        self.code = code
        self._eval_func = eval_func if eval_func else eval

    def evaluate(self, force=False):
        return self._eval_func(self.code)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.code)

    def __eq__(self, other):
        return super().__eq__(other) and self.code == other.code


class SequenceTimedCondition(SequenceItem):
    def __init__(self, *truth: List[bool]) -> None:
        super().__init__()
        self._truth = (t for t in truth)

    def evaluate(self, force=False):
        return next(self._truth)


class SequenceSuccess(SequenceItem):
    def evaluate(self, force=False):
        return True


class SequenceFailure(SequenceItem):
    def evaluate(self, force=False):
        return False


class SequenceFinish(SequenceItem):
    def evaluate(self, force=False):
        return force


# ################################################################


class SequenceSometimes(SequenceUnaryOperator):
    def __init__(self, sequence: Sequence) -> None:
        super().__init__(sequence)
        self._satisfied = False

    def evaluate(self, force=False):
        if not self._satisfied:
            if self.sequence.evaluate(force) is True:
                self._satisfied = True

        if self._satisfied:
            return True
        else:
            return False if force else None


class SequenceNegate(SequenceUnaryOperator):
    def evaluate(self, force=False):
        sequence = self.sequence.evaluate(force)
        return None if sequence is None else not sequence


class SequenceAlways(SequenceUnaryOperator):
    def __init__(self, sequence: Sequence) -> None:
        super().__init__(sequence)
        self._unsatisfied = False

    def evaluate(self, force=False):
        if not self._unsatisfied:
            sequence = self.sequence.evaluate(force)
            if sequence is False or sequence is None:
                self._unsatisfied = True

        if self._unsatisfied:
            return False
        else:
            return True if force else None


class SequenceNever(SequenceUnaryOperator):
    def __init__(self, sequence: Sequence) -> None:
        super().__init__(sequence)
        self._satisfied = False

    def evaluate(self, force=False):
        if not self._satisfied:
            sequence = self.sequence.evaluate(force)
            if sequence is True:
                self._satisfied = True

        if self._satisfied:
            return False
        else:
            return True if force else None


# ################################################################


class SequenceAnd(SequenceBinaryOperator):
    def evaluate(self, force=False):
        left = self.left.evaluate(force)
        right = self.right.evaluate(force)

        if left is False or right is False:
            return False
        elif left is True and right is True:
            return True
        else:
            return None


class SequenceOr(SequenceBinaryOperator):
    def evaluate(self, force=False):
        left = self.left.evaluate(force)
        right = self.right.evaluate(force)

        if left is True or right is True:
            return True
        elif left is False and right is False:
            return False
        else:
            return None


class SequenceThen(SequenceBinaryOperator):
    def __init__(self, left_sequence: Sequence, right_sequence: Sequence) -> None:
        super().__init__(left_sequence, right_sequence)
        self._left_satisfied = False

    def evaluate(self, force=False):
        if self._left_satisfied:
            return self.right.evaluate(force)
        else:
            self._left_satisfied = self.left.evaluate(force)
            return False if force else None


def build_sequence(expression: str, evaluation_function: Callable[[str], bool]=eval) -> Sequence:
    """
    Parse an expression and return the corresponding sequence according to the following mini-language:

     - atom:
         - "code" or 'code': a fragment of code (e.g. Python code) representing a Boolean expression that
           evaluates to true or false. The semantics is "satisfied once": as soon as the code evaluates to true once,
           the truth value of the expression remains true. This is equivalent as "sometimes 'code'" in linear
           temporal logic.
     - constants:
         - failure: this constant always evaluates to false.
         - success: this constant always evaluates to true.
     - unary operators:
         - never A: this expression evaluates to false as soon as expression A evaluates to true.
     - binary operators:
         - A and B: logical and
         - A or B: logical or
         - A -> B: this is equivalent to "(next always B) since A" in linear temporal logic, i.e. B has to be true
           (strictly) since A holds. Notice that, due to the "satisfied once" semantics of the atoms, if A and B are
           atoms, this is merely equivalent to "(A and next (sometimes B))", which means A needs to be true strictly
           before B or, in other words, A must be satisfied once, then B must be holds once.

    Keywords are case-insensitive. Parentheses can be used to group sub expressions.
    Unary operators have precedence over binary ones (e.g. "A and never B" is equivalent to "A and (never B)").
    Unary operators are right associative while binary operators are left associative (e.g. "A and B and C" is
    equivalent to "(A and B) and C").
    The binary operators are listed in decreasing priority (e.g. "A or B and C" is equivalent to "A or (B and C)",
    and "A and B -> C or D" is equivalent to "(A and B) -> (C or D)").

    Examples (assuming that expressions between quotes can be evaluated to true or false):

     - "put water" -> "put coffee": ensures water is put before coffee.
     - "put water" and "put coffee": ensures water and coffee are put. Due to the "satisfied once" semantics of the
       atoms, the order in which items are put does not matter.
     - (never "put water") or ("put water" -> "put coffee"): if water is put, then coffee must be put too.
     - never ("put water" -> "put water"): the condition will fail if water is put twice (but will succeed if water
       is put once or never put).
     - "put water" -> (never "put water"): put water exactly once.

    :param expression: an expression to parse
    :param evaluation_function: the function that will be called to evaluate nested pieces of code
    :return: a *Sequence* instance.
    """
    def unary_operator(func, term):
        term = term[0]
        return func(term[0])

    def binary_operator(func, term):
        term = term[0]
        return reduce(func, term[1:], term[0])

    condition = pyparsing.quotedString().setParseAction(
        lambda s, l, t: SequenceSometimes(SequenceCondition(t[0][1:-1], evaluation_function))
    )

    constants = (
        pyparsing.CaselessKeyword('failure').setParseAction(SequenceFailure) |
        pyparsing.CaselessKeyword('success').setParseAction(SequenceSuccess)
    )

    unary_ops = [(pyparsing.CaselessKeyword(keyword).suppress(), 1, pyparsing.opAssoc.RIGHT,
                  partial(unary_operator, func)) for keyword, func in [
        ('never', SequenceNever),
    ]]

    binary_ops = [(pyparsing.CaselessKeyword(keyword).suppress(), 2, pyparsing.opAssoc.LEFT,
                   partial(binary_operator, func)) for keyword, func in [
        ('and', SequenceAnd),
        ('or', SequenceOr),
        ('->', SequenceThen),
    ]]

    operands = (constants | condition)
    expr = pyparsing.operatorPrecedence(operands, unary_ops + binary_ops)

    try:
        return expr.parseString(expression, parseAll=True)[0]
    except pyparsing.ParseBaseException as e:
        raise StatechartError('Invalid sequential condition:\n%s' % expression) from e


from sismic.model import Statechart
import itertools
from typing import List

__all__ = ['export_to_tree']


def export_to_tree(statechart: Statechart, spaces: int=3) -> List[str]:
    """
    Provides a textual representation of the hierarchy of states belonging to
    a statechart. Only states are represented.

    :param statechart: A statechart instance
    :param spaces: spaces needed to indent a nested state
    :return: A list of lines, representation of the hierarchy of states belonging to a statechart.
    """
    def to_tree(state: str) -> List[str]:
        children = sorted(statechart.children_for(state))  # type: List[str]

        trees = map(to_tree, children)
        flat_trees = itertools.chain(*trees)
        children_repr = map(lambda x: spaces * ' ' + x, flat_trees)
        return [state] + list(children_repr)

    return to_tree(statechart.root)


import unittest
from sismic import io
from sismic.model import MacroStep, MicroStep, Transition
from sismic.interpreter import Interpreter
from sismic.stories import Story, Pause, Event
from sismic.testing import teststory_from_trace
from sismic.testing.tester import ExecutionWatcher


class StoryFromTraceTests(unittest.TestCase):
    def test_simple(self):
        with open('tests/yaml/simple.yaml') as f:
            sc = io.import_from_yaml(f)
        interpreter = Interpreter(sc)

        trace = Story([Pause(2), Event('goto s2'), Pause(3)]).tell(interpreter)
        story = teststory_from_trace(trace)

        expected = Story([
            Event('execution started'),
            Pause(2),
            Event('step started'),
            Event('state entered', state='root'),
            Event('state entered', state='s1'),
            Event('step ended'),
            Event('step started'),
            Event('event consumed', event=Event('goto s2')),
            Event('state exited', state='s1'),
            Event('transition processed', source='s1', target='s2', event=Event('goto s2')),
            Event('state entered', state='s2'),
            Event('step ended'),
            Event('step started'),
            Event('state exited', state='s2'),
            Event('transition processed', source='s2', target='s3', event=None),
            Event('state entered', state='s3'),
            Event('step ended'),
            Event('execution stopped')
        ])
        for a, b in zip(story, expected):
            self.assertEqual(a, b)
            if isinstance(a, Event):
                self.assertEqual(a.data.items(), b.data.items())
            else:
                self.assertEqual(a.duration, b.duration)


class AtomicStoryFromTraceTests(unittest.TestCase):
    def test_execution_started(self):
        story = teststory_from_trace([])
        self.assertIn(Event('execution started'), story)

        story = teststory_from_trace([MacroStep(0, [])])
        self.assertIn(Event('execution started'), story)

        story = teststory_from_trace([MacroStep(0, [MicroStep()])])
        self.assertIn(Event('execution started'), story)

    def test_execution_stopped(self):
        story = teststory_from_trace([])
        self.assertIn(Event('execution stopped'), story)

        story = teststory_from_trace([MacroStep(0, [])])
        self.assertIn(Event('execution stopped'), story)

        story = teststory_from_trace([MacroStep(0, [MicroStep()])])
        self.assertIn(Event('execution stopped'), story)

    def test_step_started(self):
        story = teststory_from_trace([])
        self.assertNotIn(Event('step started'), story)

        story = teststory_from_trace([MacroStep(0, [])])
        self.assertIn(Event('step started'), story)

        story = teststory_from_trace([MacroStep(0, [MicroStep()])])
        self.assertIn(Event('step started'), story)

    def test_step_ended(self):
        story = teststory_from_trace([])
        self.assertNotIn(Event('step ended'), story)

        story = teststory_from_trace([MacroStep(0, [])])
        self.assertIn(Event('step ended'), story)

        story = teststory_from_trace([MacroStep(0, [MicroStep()])])
        self.assertIn(Event('step ended'), story)

    def test_event_consumed(self):
        story = teststory_from_trace([])
        self.assertNotIn(Event('event consumed'), story)

        story = teststory_from_trace([MacroStep(0, [])])
        self.assertNotIn(Event('event consumed'), story)

        story = teststory_from_trace([MacroStep(0, [MicroStep()])])
        self.assertNotIn(Event('event consumed'), story)

        story = teststory_from_trace([MacroStep(0, [MicroStep(event=Event('a'))])])
        self.assertIn(Event('event consumed', event=Event('a')), story)

    def test_state_entered(self):
        story = teststory_from_trace([])
        self.assertNotIn(Event('state entered'), story)

        story = teststory_from_trace([MacroStep(0, [])])
        self.assertNotIn(Event('state entered'), story)

        story = teststory_from_trace([MacroStep(0, [MicroStep()])])
        self.assertNotIn(Event('state entered'), story)

        story = teststory_from_trace([MacroStep(0, [MicroStep(entered_states=['a'])])])
        self.assertIn(Event('state entered', state='a'), story)

        story = teststory_from_trace([MacroStep(0, [MicroStep(entered_states=['a', 'b'])])])
        self.assertIn(Event('state entered', state='a'), story)
        self.assertIn(Event('state entered', state='b'), story)
        self.assertLess(story.index(Event('state entered', state='a')),
                        story.index(Event('state entered', state='b')))

    def test_state_exited(self):
        story = teststory_from_trace([])
        self.assertNotIn(Event('state exited'), story)

        story = teststory_from_trace([MacroStep(0, [])])
        self.assertNotIn(Event('state exited'), story)

        story = teststory_from_trace([MacroStep(0, [MicroStep()])])
        self.assertNotIn(Event('state exited'), story)

        story = teststory_from_trace([MacroStep(0, [MicroStep(exited_states=['a'])])])
        self.assertIn(Event('state exited', state='a'), story)

        story = teststory_from_trace([MacroStep(0, [MicroStep(exited_states=['a', 'b'])])])
        self.assertIn(Event('state exited', state='a'), story)
        self.assertIn(Event('state exited', state='b'), story)
        self.assertLess(story.index(Event('state exited', state='a')),
                        story.index(Event('state exited', state='b')))

    def test_transition_processed(self):
        story = teststory_from_trace([])
        self.assertNotIn(Event('transition processed'), story)

        story = teststory_from_trace([MacroStep(0, [])])
        self.assertNotIn(Event('transition processed'), story)

        story = teststory_from_trace([MacroStep(0, [MicroStep()])])
        self.assertNotIn(Event('transition processed'), story)

        story = teststory_from_trace([MacroStep(0, [MicroStep(event=Event('c'), transition=Transition('a', 'b', event='c'))])])
        self.assertIn(Event('transition processed', source='a', target='b', event=Event('c')), story)

    def test_event_sent(self):
        story = teststory_from_trace([])
        self.assertNotIn(Event('event sent'), story)

        story = teststory_from_trace([MacroStep(0, [])])
        self.assertNotIn(Event('event sent'), story)

        story = teststory_from_trace([MacroStep(0, [MicroStep()])])
        self.assertNotIn(Event('event sent'), story)

        story = teststory_from_trace([MacroStep(0, [MicroStep(sent_events=[Event('a')])])])
        self.assertIn(Event('event sent', event=Event('a')), story)

        story = teststory_from_trace([MacroStep(0, [MicroStep(sent_events=[Event('a'), Event('b')])])])
        self.assertIn(Event('event sent', event=Event('a')), story)
        self.assertIn(Event('event sent', event=Event('b')), story)
        self.assertLess(story.index(Event('event sent', event=Event('a'))),
                       story.index(Event('event sent', event=Event('b'))))


class ElevatorStoryTests(unittest.TestCase):
    def setUp(self):
        with open('docs/examples/elevator/elevator.yaml') as f:
            sc = io.import_from_yaml(f)
        self.tested = Interpreter(sc)

    def test_7th_floor_never_reached(self):
        story = Story([Event('floorSelected', floor=8)])
        trace = story.tell(self.tested)  # self.tested is an interpreter for our elevator

        test_story = teststory_from_trace(trace)

        with open('docs/examples/elevator/tester_elevator_7th_floor_never_reached.yaml') as f:
            tester = Interpreter(io.import_from_yaml(f))
        test_story.tell(tester)
        self.assertFalse(tester.final)

    def test_7th_floor_never_reached_fails(self):
        story = Story([Event('floorSelected', floor=4), Pause(2), Event('floorSelected', floor=7)])
        trace = story.tell(self.tested)  # self.tested is an interpreter for our elevator

        test_story = teststory_from_trace(trace)

        with open('docs/examples/elevator/tester_elevator_7th_floor_never_reached.yaml') as f:
            tester = Interpreter(io.import_from_yaml(f))
        test_story.tell(tester)
        self.assertTrue(tester.final)

    def test_elevator_moves_after_10s(self):
        stories = [
            Story([Event('floorSelected', floor=4)]),
            Story([Event('floorSelected', floor=0)]),
            Story([Event('floorSelected', floor=4), Pause(10)]),
            Story([Event('floorSelected', floor=0), Pause(10)]),
            Story([Event('floorSelected', floor=4), Pause(9)]),
            Story([Event('floorSelected', floor=0), Pause(9)]),
        ]

        for story in stories:
            with self.subTest(story=story):
                # Reopen because we need to reset it
                with open('docs/examples/elevator/elevator.yaml') as f:
                    sc = io.import_from_yaml(f)
                tested = Interpreter(sc)

                test_story = teststory_from_trace(story.tell(tested))

                with open('docs/examples/elevator/tester_elevator_moves_after_10s.yaml') as f:
                    tester = Interpreter(io.import_from_yaml(f))
                test_story.tell(tester)
                self.assertFalse(tester.final)


class WatchElevatorTests(unittest.TestCase):
    def setUp(self):
        with open('docs/examples/elevator/elevator.yaml') as f:
            sc = io.import_from_yaml(f)
        self.tested = Interpreter(sc)
        self.watcher = ExecutionWatcher(self.tested)

    def test_7th_floor_never_reached(self):
        with open('docs/examples/elevator/tester_elevator_7th_floor_never_reached.yaml') as f:
            tester_sc = io.import_from_yaml(f)

        tester = self.watcher.watch_with(tester_sc)
        self.watcher.start()

        # Send elevator to 4th
        self.tested.queue(Event('floorSelected', floor=4)).execute()
        self.watcher.stop()
        self.assertFalse(tester.final)

    def test_7th_floor_never_reached_fails(self):
        with open('docs/examples/elevator/tester_elevator_7th_floor_never_reached.yaml') as f:
            tester_sc = io.import_from_yaml(f)

        tester = self.watcher.watch_with(tester_sc)
        self.watcher.start()

        # Send elevator to 7th
        self.tested.queue(Event('floorSelected', floor=7)).execute()
        self.watcher.stop()
        self.assertTrue(tester.final)


    def test_7th_floor_never_reached_fails_fast(self):
        with open('docs/examples/elevator/tester_elevator_7th_floor_never_reached.yaml') as f:
            tester_sc = io.import_from_yaml(f)

        tester = self.watcher.watch_with(tester_sc, fails_fast=True)
        self.watcher.start()

        # Send elevator to 7th
        with self.assertRaises(AssertionError) as cm:
            self.tested.queue(Event('floorSelected', floor=7)).execute()
        self.assertTrue(tester.final)


    def test_destination_reached(self):
        with open('docs/examples/elevator/tester_elevator_destination_reached.yaml') as f:
            tester_statechart = io.import_from_yaml(f)

        # Create the interpreter and the watcher
        watcher = ExecutionWatcher(self.tested)

        # Add the tester and start watching
        tester = watcher.watch_with(tester_statechart)
        watcher.start()

        # Send the elevator to 4th
        self.tested.queue(Event('floorSelected', floor=4)).execute(max_steps=2)
        self.assertEqual(tester.context['destinations'], [4])

        self.tested.execute()
        self.assertEqual(tester.context['destinations'], [])

        # Stop watching. The tester must be in a final state
        watcher.stop()

        self.assertFalse(tester.final)


from itertools import combinations

from collections import deque

from sismic import model
from sismic.code import Evaluator, PythonEvaluator
from sismic.exceptions import InvariantError, PreconditionError, PostconditionError, SequentialConditionError
from sismic.exceptions import NonDeterminismError, ConflictingTransitionsError
from typing import Optional, List, Dict, Set, Union, Callable, Any, cast, Iterable, Mapping

__all__ = ['Interpreter']


class Interpreter:
    """
    A discrete interpreter that executes a statechart according to a semantic close to SCXML.

    :param statechart: statechart to interpret
    :param evaluator_klass: An optional callable (eg. a class) that takes an interpreter and an optional initial
        context as input and return an *Evaluator* instance that will be used to initialize the interpreter.
        By default, the *PythonEvaluator* class will be used.
    :param initial_context: an optional initial context that will be provided to the evaluator.
        By default, an empty context is provided
    :param ignore_contract: set to True to ignore contract checking during the execution.
    """

    def __init__(self, statechart: model.Statechart, *,
                 evaluator_klass: Callable[['Interpreter'], Evaluator]=PythonEvaluator,
                 initial_context: Mapping[str, Any]=None,
                 ignore_contract: bool=False) -> None:
        # Internal variables
        self._ignore_contract = ignore_contract
        self._statechart = statechart

        self._initialized = False

        # Internal clock
        self._time = 0  # type: float

        # History states memory
        self._memory = {}  # type: Dict[str, Optional[List[str]]]

        # Set of active states
        self._configuration = set()  # type: Set[str]

        # External events queue
        self._external_events = deque()  # type: deque[model.Event]

        # Internal events queue
        self._internal_events = deque()  # type: deque[model.InternalEvent]

        # Bound callables
        self._bound = []  # type: List[Callable[[model.Event], Any]]

        # Evaluator
        self._evaluator = evaluator_klass(self, initial_context=initial_context)  # type: ignore
        for event in self._evaluator.execute_statechart(statechart):
            self.raise_event(event)

    @property
    def time(self) -> float:
        """
        Time value (in seconds) for the internal clock
        """
        return self._time

    @time.setter
    def time(self, value: float):
        """
        Set the time of the internal clock

        :param value: time value (in seconds)
        """
        self._time = value

    @property
    def configuration(self) -> List[str]:
        """
        List of active states names, ordered by depth. Ties are broken according to the lexicographic order
        on the state name.
        """
        return sorted(self._configuration, key=lambda s: (self._statechart.depth_for(s), s))

    @property
    def context(self) -> Mapping[str, Any]:
        """
        The context of execution.
        """
        return self._evaluator.context

    @property
    def final(self) -> bool:
        """
        Boolean indicating whether this interpreter is in a final configuration.
        """
        return self._initialized and len(self._configuration) == 0

    @property
    def statechart(self) -> model.Statechart:
        """
        Embedded statechart
        """
        return self._statechart

    def bind(self, interpreter_or_callable: Union['Interpreter', Callable[[model.Event], Any]]) -> 'Interpreter':
        """
        Bind an interpreter or a callable to the current interpreter.
        Each time an internal event is sent by this interpreter, any bound object will be called
        with the same event. If *interpreter_or_callable* is an *Interpreter* instance,  its *queue* method is called.
        This is, if *i1* and *i2* are interpreters, *i1.bind(i2)* is equivalent to *i1.bind(i2.queue)*.

        :param interpreter_or_callable: interpreter or callable to bind
        :return: *self* so it can be chained
        """
        if isinstance(interpreter_or_callable, Interpreter):
            bound_callable = interpreter_or_callable.queue
        else:
            bound_callable = interpreter_or_callable

        self._bound.append(bound_callable)
        return self

    def raise_event(self, event: model.Event) -> None:
        """
        Raise an event from the statechart.
        Events are propagated to bound interpreters as non-internal events, and added to the internal queue of the
        current interpreter.

        :param event: raised event.
        """
        if isinstance(event, model.InternalEvent):
            # Propagate event to bound callable as an "external" event
            external_event = model.Event(event.name, **event.data)
            for bound_callable in self._bound:
                bound_callable(external_event)

            # Add to current interpreter's internal queue
            self._internal_events.append(event)
        else:
            raise ValueError('Only InternalEvent instances are supported, not {}.'.format(type(event)))

    def queue(self, event: model.Event) -> 'Interpreter':
        """
        Queue an event to the interpreter.

        :param event: an *Event* instance.
        :return: *self* so it can be chained.
        """
        if isinstance(event, model.Event):
            self._external_events.append(event)
        else:
            raise ValueError('{} is not an Event instance'.format(event))
        return self

    def execute(self, max_steps: int=-1) -> List[model.MacroStep]:
        """
        Repeatedly calls *execute_once* and return a list containing
        the returned values of *execute_once*.

        Notice that this does NOT return an iterator but computes the whole list first
        before returning it.

        :param max_steps: An upper bound on the number steps that are computed and returned.
            Default is -1, no limit. Set to a positive integer to avoid infinite loops
            in the statechart execution.
        :return: A list of *MacroStep* instances
        """
        returned_steps = []
        i = 0
        macro_step = self.execute_once()
        while macro_step:
            returned_steps.append(macro_step)
            i += 1
            if 0 < max_steps == i:
                break
            macro_step = self.execute_once()
        return returned_steps

    def execute_once(self) -> model.MacroStep:
        """
        Processes a transition based on the oldest queued event (or no event if an eventless transition
        can be processed), and stabilizes the interpreter in a stable situation (ie. processes initial states,
        history states, etc.). When multiple transitions are selected, they are atomically processed:
        states are exited, transition is processed, states are entered, statechart is stabilized and only
        after that, the next transition is processed.

        :return: a macro step or *None* if nothing happened
        """
        event = None  # type: Optional[model.Event]

        # Initialization
        if not self._initialized:
            computed_steps = [model.MicroStep(entered_states=[self._statechart.root])]
            self._initialized = True
        else:
            # Look for eventless transitions first
            transitions = self._select_transitions(event=None)
            if len(transitions) == 0:
                # Look for evented transitions
                event = self._select_event()
                if event is None:
                    return None  # No event means no step!
                transitions = self._select_transitions(event=event)

            # No transition? Empty step!
            if len(transitions) == 0:
                computed_steps = [model.MicroStep(event=event)]
            else:
                # Select the transitions that will be performed
                transitions = self._sort_transitions(
                    self._filter_transitions(transitions)
                )
                computed_steps = self._create_steps(event, transitions)

        # Execute the steps
        self._evaluator.on_step_starts(event)

        executed_steps = []
        for step in computed_steps:
            executed_steps.append(self._apply_step(step))
            executed_steps.extend(self._stabilize())

        macro_step = model.MacroStep(time=self.time, steps=executed_steps)

        # Check state invariants
        configuration = self.configuration  # Use self.configuration to benefit from the sorting by depth
        for name in configuration:
            state = self._statechart.state_for(name)
            self._evaluate_contract_conditions(state, 'invariants', macro_step)

        # Check state sequential conditions
        if not self._ignore_contract:
            for name in configuration:
                state = self._statechart.state_for(name)
                for condition in self._evaluator.update_sequential_conditions(state):
                    raise SequentialConditionError(configuration=self.configuration, step=macro_step, obj=state,
                                                   assertion=condition, context=self.context)

        return macro_step

    def _select_event(self) -> Optional[model.Event]:
        """
        Return (and consume!) the next available event if any.
        This method prioritizes internal events over external ones.

        :return: An instance of Event or None if no event is available
        """
        # Internal events are processed first
        if len(self._internal_events) > 0:
            return self._internal_events.popleft()
        elif len(self._external_events) > 0:
            return self._external_events.popleft()
        else:
            return None

    def _select_transitions(self, event: model.Event=None) -> List[model.Transition]:
        """
        Return a list of transitions that can be triggered according to the given event, or eventless
        transition if *event* is None.

        :param event: event to consider
        :return: a list of *Transition* instances
        """
        transitions = []

        # Retrieve the firable transitions for all active state
        for transition in self._statechart.transitions:
            if (transition.event == getattr(event, 'name', None) and transition.source in self._configuration and
                    (transition.guard is None or self._evaluator.evaluate_guard(transition, event))):
                transitions.append(transition)
        return transitions

    def _filter_transitions(self, transitions: List[model.Transition]) -> List[model.Transition]:
        """
        Given a list of transitions, return a filtered list of transitions with respect to the
        inner-first/source-state semantic.

        :param transitions: a list of *Transition* instances
        :return: a list of *Transition* instances
        """
        removed_transitions = set()
        for transition in transitions:
            source_state_descendants = self._statechart.descendants_for(transition.source)
            for other_transition in transitions:
                if other_transition.source in source_state_descendants:
                    removed_transitions.add(transition)
                    break

        return list(set(transitions).difference(removed_transitions))

    def _sort_transitions(self, transitions: List[model.Transition]) -> List[model.Transition]:
        """
        Given a list of triggered transitions, return a list of transitions in an order that represents
        the order in which they have to be processed.

        :param transitions: a list of *Transition* instances
        :return: an ordered list of *Transition* instances
        :raise ExecutionError: In case of non-determinism (*NonDeterminismError*) or conflicting
            transitions (*ConflictingTransitionsError*).
        """
        if len(transitions) > 1:
            # If more than one transition, we check (1) they are from separate regions and (2) they do not conflict
            # Two transitions conflict if one of them leaves the parallel state
            for t1, t2 in combinations(transitions, 2):
                # Check (1)
                lca = self._statechart.least_common_ancestor(t1.source, t2.source)
                lca_state = self._statechart.state_for(lca)

                # Their LCA must be an orthogonal state!
                if not isinstance(lca_state, model.OrthogonalState):
                    raise NonDeterminismError(
                        'Non-determinist choice between transitions {t1} and {t2}'
                        '\nConfiguration is {c}\nEvent is {e}\nTransitions are:{t}\n'
                        .format(c=self.configuration, e=t1.event, t=transitions, t1=t1, t2=t2)
                    )

                # Check (2)
                # This check must be done wrt. to LCA, as the combination of from_states could
                # come from nested parallel regions!
                for transition in [t1, t2]:
                    last_before_lca = transition.source
                    for state in self._statechart.ancestors_for(transition.source):
                        if state == lca:
                            break
                        last_before_lca = state
                    # Target must be a descendant (or self) of this state
                    if (transition.target and
                            (transition.target not in
                             [last_before_lca] + self._statechart.descendants_for(last_before_lca))):
                        raise ConflictingTransitionsError(
                            'Conflicting transitions: {t1} and {t2}'
                            '\nConfiguration is {c}\nEvent is {e}\nTransitions are:{t}\n'
                            .format(c=self.configuration, e=t1.event, t=transitions, t1=t1, t2=t2)
                        )

            # Define an arbitrary order based on the depth and the name of source states.
            transitions = sorted(transitions, key=lambda t: (-self._statechart.depth_for(t.source), t.source))

        return transitions

    def _create_steps(self, event: model.Event,
                      transitions: Iterable[model.Transition]) -> List[model.MicroStep]:
        """
        Return a (possibly empty) list of micro steps. Each micro step corresponds to the process of a transition
        matching given event.

        :param event: the event to consider, if any
        :param transitions: the transitions that should be processed
        :return: a list of micro steps.
        """
        returned_steps = []
        for transition in transitions:
            # Internal transition
            if transition.target is None:
                returned_steps.append(model.MicroStep(event=event, transition=transition))
                continue

            lca = self._statechart.least_common_ancestor(transition.source, transition.target)
            from_ancestors = self._statechart.ancestors_for(transition.source)
            to_ancestors = self._statechart.ancestors_for(transition.target)

            # Exited states
            exited_states = []

            # last_before_lca is the "highest" ancestor of from_state that is a child of LCA
            last_before_lca = transition.source
            for state in from_ancestors:
                if state == lca:
                    break
                last_before_lca = state

            # Take all the descendants of this state and list the ones that are active
            for descendant in self._statechart.descendants_for(last_before_lca)[::-1]:  # Mind the reversed order!
                # Only leave states that are currently active
                if descendant in self._configuration:
                    exited_states.append(descendant)

            # Add last_before_lca as it is a child of LCA that must be exited
            if last_before_lca in self._configuration:
                exited_states.append(last_before_lca)

            # Entered states
            entered_states = [transition.target]
            for state in to_ancestors:
                if state == lca:
                    break
                entered_states.insert(0, state)

            returned_steps.append(model.MicroStep(event=event, transition=transition,
                                                  entered_states=entered_states, exited_states=exited_states))

        return returned_steps

    def _create_stabilization_step(self, names: Iterable[str]) -> model.MicroStep:
        """
        Return a stabilization step, ie. a step that lead to a more stable situation
        for the current statechart. Stabilization means:

         - Enter the initial state of a compound state with no active child
         - Enter the memory of a history state
         - Enter the children of an orthogonal state with no active child
         - Exit active states if all "deepest" (leaves) states are final

        :param names: List of states to consider (usually, the active configuration)
        :return: A *MicroStep* instance or *None* if this statechart can not be more stabilized
        """
        # Check if we are in a set of "stable" states
        leaves_names = self._statechart.leaf_for(names)
        leaves = sorted(map(self._statechart.state_for, leaves_names),
                        key=lambda s: (-self._statechart.depth_for(s.name), s.name))

        # Final states?
        if len(leaves) > 0 and all([isinstance(s, model.FinalState) for s in leaves]):
            # Leave all states
            exited_states = sorted(self._configuration, key=lambda s: (-self._statechart.depth_for(s), s))
            return model.MicroStep(exited_states=exited_states)

        # Otherwise, develop history, compound and orthogonal states.
        for leaf in leaves:
            name = cast(model.StateMixin, leaf).name
            if isinstance(leaf, model.HistoryStateMixin):
                states_to_enter = self._memory.get(name, [leaf.memory])
                states_to_enter.sort(key=lambda x: (self._statechart.depth_for(x), x))
                return model.MicroStep(entered_states=states_to_enter, exited_states=[name])
            elif isinstance(leaf, model.OrthogonalState) and self._statechart.children_for(leaf.name):
                return model.MicroStep(entered_states=sorted(self._statechart.children_for(leaf.name)))
            elif isinstance(leaf, model.CompoundState) and leaf.initial:
                return model.MicroStep(entered_states=[leaf.initial])

    def _apply_step(self, step: model.MicroStep) -> model.MicroStep:
        """
        Apply given *MicroStep* on this statechart

        :param step: *MicroStep* instance
        :return: a new MicroStep, completed with sent events
        """
        entered_states = list(map(self._statechart.state_for, step.entered_states))
        exited_states = list(map(self._statechart.state_for, step.exited_states))

        active_configuration = set(self._configuration)  # Copy

        sent_events = []  # type: List[model.Event]

        # Exit states
        for state in exited_states:
            # Sequential conditions
            if not self._ignore_contract:
                for condition in self._evaluator.evaluate_sequential_conditions(state):
                    raise SequentialConditionError(configuration=self.configuration, step=step, obj=state,
                                                   assertion=condition, context=self.context)

            # Execute exit action
            sent_events.extend(self._evaluator.execute_onexit(state))

            # Postconditions
            self._evaluate_contract_conditions(state, 'postconditions', step)

            # Deal with history
            if isinstance(state, model.CompoundState):
                # Look for an HistoryStateMixin among its children
                for child_name in self._statechart.children_for(state.name):
                    child = self._statechart.state_for(child_name)
                    if isinstance(child, model.DeepHistoryState):
                        # This MUST contain at least one element!
                        active = active_configuration.intersection(self._statechart.descendants_for(state.name))
                        assert len(active) >= 1
                        self._memory[child.name] = list(active)
                    elif isinstance(child, model.ShallowHistoryState):
                        # This MUST contain exactly one element!
                        active = active_configuration.intersection(self.statechart.children_for(state.name))
                        assert len(active) == 1
                        self._memory[child.name] = list(active)

            # Remove state from active configuration
            self._configuration.remove(state.name)

        # Execute transition
        if step.transition:
            # Preconditions and invariants
            self._evaluate_contract_conditions(step.transition, 'preconditions', step)
            self._evaluate_contract_conditions(step.transition, 'invariants', step)

            sent_events.extend(self._evaluator.execute_action(step.transition, step.event))

            # Postconditions and invariants
            self._evaluate_contract_conditions(step.transition, 'postconditions', step)
            self._evaluate_contract_conditions(step.transition, 'invariants', step)

        # Enter states
        for state in entered_states:
            # Preconditions
            self._evaluate_contract_conditions(state, 'preconditions', step)

            # Execute entry action
            sent_events.extend(self._evaluator.execute_onentry(state))

            # Sequential conditions
            if not self._ignore_contract:
                self._evaluator.initialize_sequential_conditions(state)
                for condition in self._evaluator.update_sequential_conditions(state):
                    raise SequentialConditionError(configuration=self.configuration, step=step, obj=state,
                                                   assertion=condition, context=self.context)

            # Update configuration
            self._configuration.add(state.name)

        # Send events
        for event in sent_events:
            self.raise_event(event)

        return model.MicroStep(event=step.event, transition=step.transition,
                               entered_states=step.entered_states, exited_states=step.exited_states,
                               sent_events=sent_events)

    def _stabilize(self) -> List[model.MicroStep]:
        """
        Compute, apply and return stabilization steps.

        :return: A list of applied  *MicroStep* instances,
        """
        # Stabilization
        steps = []
        step = self._create_stabilization_step(self._configuration)
        while step is not None:
            steps.append(self._apply_step(step))
            step = self._create_stabilization_step(self._configuration)
        return steps

    def _evaluate_contract_conditions(self, obj: Union[model.Transition, model.StateMixin],
                                      cond_type: str,
                                      step: Union[model.MacroStep, model.MicroStep]=None) -> None:
        """
        Evaluate the conditions for given object.

        :param obj: object with preconditions, postconditions or invariants
        :param cond_type: either "preconditions", "postconditions" or "invariants"
        :param step: step in which the check occurs.
        :raises ContractError: if a condition fails and *ignore_contract* is False.
        """
        if self._ignore_contract:
            return

        exception_klass = cast(Callable[..., Exception], {'preconditions': PreconditionError,
                                                          'postconditions': PostconditionError,
                                                          'invariants': InvariantError}[cond_type])

        unsatisfied_conditions = getattr(self._evaluator, 'evaluate_' + cond_type)(obj, getattr(step, 'event', None))

        for condition in unsatisfied_conditions:
            raise exception_klass(configuration=self.configuration, step=step, obj=obj,
                                  assertion=condition, context=self.context)

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self._statechart)




import unittest
from sismic import io
from sismic.interpreter import Interpreter
from sismic.interpreter.helpers import run_in_background, log_trace, coverage_from_trace
from sismic import exceptions
from sismic.code import DummyEvaluator
from sismic.model import Event, InternalEvent, MacroStep, MicroStep, Transition

from collections import Counter


class LogTraceTests(unittest.TestCase):
    def setUp(self):
        with open('docs/examples/elevator/elevator.yaml') as f:
            sc = io.import_from_yaml(f)
        self.tested = Interpreter(sc)
        self.steps = log_trace(self.tested)

    def test_empty_trace(self):
        self.assertEqual(self.steps, [])

    def test_nonempty_trace(self):
        self.tested.queue(Event('floorSelected', floor=4)).execute()
        self.assertTrue(len(self.steps) > 0)

    def test_log_content(self):
        self.tested.queue(Event('floorSelected', floor=4))
        steps = self.tested.execute()
        self.assertSequenceEqual(self.steps, steps)


class RunInBackgroundTests(unittest.TestCase):
    def test_run_in_background(self):
        with open('tests/yaml/simple.yaml') as f:
            sc = io.import_from_yaml(f)
        interpreter = Interpreter(sc)
        task = run_in_background(interpreter, 0.001)
        interpreter.queue(Event('goto s2'))
        interpreter.queue(Event('goto final'))
        task.join()
        self.assertTrue(interpreter.final)


class CoverageFromTraceTests(unittest.TestCase):
    def test_empty_trace(self):
        self.assertEqual(coverage_from_trace([]), (Counter(), Counter()))

    def test_single_step(self):
        trace = [MacroStep(0, steps=[
            MicroStep(entered_states=['a', 'b', 'c'], transition=Transition('x')),
            MicroStep(entered_states=['a', 'b'], transition=Transition('x')),
            MicroStep(entered_states=['a']),
            MicroStep(entered_states=[])
        ])]
        self.assertEqual(coverage_from_trace(trace), (Counter(a=3, b=2, c=1), Counter({Transition('x'): 2})))

    def test_multiple_steps(self):
        trace = [MacroStep(0, steps=[
            MicroStep(entered_states=['a', 'b', 'c'], transition=Transition('x')),
            MicroStep(entered_states=['a', 'b'], transition=Transition('x')),
            MicroStep(entered_states=['a']),
            MicroStep(entered_states=[])
        ])]

        trace.extend(trace)
        self.assertEqual(coverage_from_trace(trace), (Counter(a=6, b=4, c=2), Counter({Transition('x'): 4})))


class SimulatorSimpleTests(unittest.TestCase):
    def setUp(self):
        with open('tests/yaml/simple.yaml') as f:
            sc = io.import_from_yaml(f)
        self.interpreter = Interpreter(sc, evaluator_klass=DummyEvaluator)
        # Stabilization
        self.interpreter.execute_once()

    def test_init(self):
        self.assertEqual(self.interpreter.configuration, ['root', 's1'])
        self.assertFalse(self.interpreter.final)

    def test_queue(self):
        self.interpreter.queue(Event('e1'))
        self.assertEqual(self.interpreter._select_event(), Event('e1'))

        self.interpreter.queue(InternalEvent('e1'))
        self.assertEqual(self.interpreter._select_event(), InternalEvent('e1'))

        with self.assertRaises(ValueError):
            self.interpreter.queue('e1')

    def test_simple_configuration(self):
        self.interpreter.execute_once()  # Should do nothing!
        self.assertEqual(self.interpreter.configuration, ['root', 's1'])
        self.interpreter.queue(Event('goto s2'))
        self.interpreter.execute_once()
        self.assertEqual(self.interpreter.configuration, ['root', 's2'])
        self.interpreter.execute_once()
        self.assertEqual(self.interpreter.configuration, ['root', 's3'])

    def test_simple_entered(self):
        self.interpreter.queue(Event('goto s2'))
        self.assertEqual(self.interpreter.execute_once().entered_states, ['s2'])
        self.interpreter.queue(Event('goto final'))
        self.assertEqual(self.interpreter.execute_once().entered_states, ['s3'])
        self.assertEqual(self.interpreter.execute_once().entered_states, ['final'])
        self.assertEqual(self.interpreter.configuration, [])
        self.assertTrue(self.interpreter.final)

    def test_simple_final(self):
        self.interpreter.queue(Event('goto s2')).queue(Event('goto final'))
        self.interpreter.execute()
        self.assertTrue(self.interpreter.final)


class InternalTests(unittest.TestCase):
    def setUp(self):
        with open('tests/yaml/internal.yaml') as f:
            self.sc = io.import_from_yaml(f)
        self.interpreter = Interpreter(self.sc)
        # Stabilization
        self.interpreter.execute_once()

    def testInternalSent(self):
        step = self.interpreter.execute_once()
        self.assertEqual(step.event.name, 'next')

    def testInternalBeforeExternal(self):
        self.interpreter.queue(Event('not_next'))
        step = self.interpreter.execute_once()
        self.assertEqual(step.event.name, 'next')

        step = self.interpreter.execute_once()
        self.assertEqual(step.event, None)
        self.assertEqual(step.entered_states, ['s2'])

        step = self.interpreter.execute_once()
        self.assertEqual(step.event.name, 'not_next')

    def testActiveGuard(self):
        self.interpreter.execute()
        self.assertTrue(self.interpreter.final)


class SimulatorNonDeterminismTests(unittest.TestCase):
    def test_nondeterminism(self):
        with open('tests/yaml/nondeterministic.yaml') as f:
            sc = io.import_from_yaml(f)
        interpreter = Interpreter(sc, evaluator_klass=DummyEvaluator)
        # Stabilization
        interpreter.execute_once()

        with self.assertRaises(exceptions.NonDeterminismError):
            interpreter.execute_once()


class SimulatorHistoryTests(unittest.TestCase):
    def setUp(self):
        with open('tests/yaml/history.yaml') as f:
            sc = io.import_from_yaml(f)
        self.interpreter = Interpreter(sc, evaluator_klass=DummyEvaluator)
        # Stabilization
        self.interpreter.execute_once()

    def test_memory(self):
        self.interpreter.queue(Event('next')).execute_once()
        self.assertEqual(sorted(self.interpreter.configuration), ['loop', 'root', 's2'])

        step = self.interpreter.queue(Event('pause')).execute_once()
        self.assertEqual(step.exited_states, ['s2', 'loop'])
        self.assertEqual(sorted(self.interpreter.configuration), ['pause', 'root'])

    def test_resume_memory(self):
        self.interpreter.queue(Event('next')).queue(Event('pause')).queue(Event('continue'))
        steps = self.interpreter.execute()
        step = steps[-1]

        self.assertEqual(step.entered_states, ['loop', 'loop.H', 's2'])
        self.assertEqual(step.exited_states, ['pause', 'loop.H'])
        self.assertEqual(sorted(self.interpreter.configuration), ['loop', 'root', 's2'])

    def test_after_memory(self):
        self.interpreter.queue(Event('next')).queue(Event('pause')).queue(Event('continue'))
        self.interpreter.queue(Event('next')).queue(Event('next'))
        self.interpreter.execute()
        self.assertEqual(sorted(self.interpreter.configuration), ['loop', 'root', 's1'])

        self.interpreter.queue(Event('pause')).queue(Event('stop'))
        self.interpreter.execute()
        self.assertTrue(self.interpreter.final)


class SimulatorDeepHistoryTests(unittest.TestCase):
    def setUp(self):
        with open('tests/yaml/deep_history.yaml') as f:
            sc = io.import_from_yaml(f)
        self.interpreter = Interpreter(sc, evaluator_klass=DummyEvaluator)
        # Stabilization
        self.interpreter.execute_once()

    def test_deep_memory(self):
        self.interpreter.queue(Event('next1')).queue(Event('next2'))
        self.interpreter.execute()
        self.assertEqual(sorted(self.interpreter.configuration), ['active', 'concurrent_processes', 'process_1', 'process_2', 'root', 's12', 's22'])

        self.interpreter.queue(Event('error1'))
        self.interpreter.execute()
        self.assertEqual(self.interpreter.configuration, ['root', 'pause'])
        self.assertEqual(sorted(self.interpreter._memory['active.H*']), ['concurrent_processes', 'process_1', 'process_2', 's12', 's22'])

        self.interpreter.queue(Event('continue'))
        self.interpreter.execute()
        self.assertEqual(sorted(self.interpreter.configuration), ['active', 'concurrent_processes', 'process_1',
                                                           'process_2', 'root', 's12', 's22'])

    def test_entered_order(self):
        self.interpreter.queue(Event('next1')).queue(Event('next2')).queue(Event('pause'))
        step = self.interpreter.execute()[-1]

        self.assertEqual(step.entered_states, ['pause'])
        self.assertEqual(sorted(self.interpreter.configuration), ['pause', 'root'])

        step = self.interpreter.queue(Event('continue')).execute_once()
        self.assertTrue(step.entered_states.index('active') < step.entered_states.index('active.H*'))
        self.assertTrue(step.entered_states.index('active.H*') < step.entered_states.index('concurrent_processes'))
        self.assertTrue(step.entered_states.index('concurrent_processes') < step.entered_states.index('process_1'))
        self.assertTrue(step.entered_states.index('concurrent_processes') < step.entered_states.index('process_2'))
        self.assertTrue(step.entered_states.index('process_1') < step.entered_states.index('s12'))
        self.assertTrue(step.entered_states.index('process_2') < step.entered_states.index('s22'))

        self.interpreter.queue(Event('next1')).queue(Event('next2')).execute()
        self.assertTrue(self.interpreter.final)

    def test_exited_order(self):
        self.interpreter.queue(Event('next1')).queue(Event('next2')).queue(Event('pause'))
        step = self.interpreter.execute()[-1]

        self.assertEqual(step.exited_states, ['s12', 's22', 'process_1', 'process_2', 'concurrent_processes', 'active'])
        self.assertEqual(sorted(self.interpreter.configuration), ['pause', 'root'])

        step = self.interpreter.queue(Event('continue')).execute_once()
        self.assertEqual(step.exited_states, ['pause', 'active.H*'])

        self.interpreter.queue(Event('next1')).queue(Event('next2')).execute()
        self.assertTrue(self.interpreter.final)


class InfiniteExecutionTests(unittest.TestCase):
    def setUp(self):
        with open('tests/yaml/infinite.yaml') as f:
            sc = io.import_from_yaml(f)
        self.interpreter = Interpreter(sc)
        # Stabilization
        self.interpreter.execute_once()

    def test_three_steps(self):
        self.assertEqual(self.interpreter.configuration, ['root', 's1'])
        self.interpreter.execute_once()
        self.assertEqual(self.interpreter.configuration, ['root', 's2'])
        self.interpreter.execute_once()
        self.assertEqual(self.interpreter.configuration, ['root', 's1'])
        self.interpreter.execute_once()
        self.assertEqual(self.interpreter.configuration, ['root', 's2'])
        self.assertEqual(self.interpreter.context['x'], 2)  # x is incremented in s1.on_entry

    def test_auto_three_steps(self):
        self.interpreter.execute(max_steps=3)

        self.assertEqual(self.interpreter.configuration, ['root', 's2'])
        self.assertEqual(self.interpreter.context['x'], 2)  # x is incremented in s1.on_entry

    def test_auto_stop(self):
        self.interpreter.execute()

        self.assertTrue(self.interpreter.final)
        self.assertEqual(self.interpreter.context['x'], 100)


class ParallelExecutionTests(unittest.TestCase):
    def setUp(self):
        with open('tests/yaml/parallel.yaml') as f:
            sc = io.import_from_yaml(f)
        self.interpreter = Interpreter(sc)
        # Stabilization
        self.interpreter.execute_once()

    def test_concurrent_transitions(self):
        step = self.interpreter.queue(Event('nextA')).execute_once()

        self.assertEqual(self.interpreter.configuration, ['root', 's1', 'p1', 'p2', 'a1', 'a2'])
        self.assertLess(step.exited_states.index('initial1'), step.exited_states.index('initial2'))
        self.assertLess(step.entered_states.index('a1'), step.entered_states.index('a2'))

    def test_concurrent_transitions_nested_target(self):
        self.interpreter.queue(Event('nextA')).queue(Event('reset1'))
        self.interpreter.execute()

        self.assertEqual(self.interpreter.configuration, ['root', 's1', 'p1', 'p2', 'a2', 'initial1'])

    def test_unnested_transitions(self):
        self.interpreter.queue(Event('nextA')).queue(Event('nextA'))
        self.interpreter.execute()

        self.assertEqual(self.interpreter.configuration, ['root', 's1', 'p1', 'p2', 'a2', 'initial1'])

    def test_unnested_transitions_2(self):
        self.interpreter.queue(Event('nextA')).queue(Event('nextB'))
        self.interpreter.execute()

        self.assertEqual(self.interpreter.configuration, ['root', 's1', 'p1', 'p2', 'b1', 'b2'])

    def test_conflicting_transitions(self):
        self.interpreter.queue(Event('nextA')).queue(Event('nextB')).queue(Event('conflict1'))
        self.interpreter.execute_once()
        self.interpreter.execute_once()

        with self.assertRaises(exceptions.ConflictingTransitionsError):
            self.interpreter.execute_once()

    def test_conflicting_transitions_2(self):
        self.interpreter.queue(Event('nextA')).queue(Event('nextB')).queue(Event('conflict2'))
        self.interpreter.execute_once()
        self.interpreter.execute_once()

        with self.assertRaises(exceptions.ConflictingTransitionsError):
            self.interpreter.execute_once()


class NestedParallelExecutionTests(unittest.TestCase):
    def setUp(self):
        with open('tests/yaml/nested_parallel.yaml') as f:
            sc = io.import_from_yaml(f)
        self.interpreter = Interpreter(sc)
        # Stabilization
        self.interpreter.execute_once()
        self.common_states = ['root', 's1', 'p1', 'p2', 'r1', 'r2', 'r3', 'r4']

    def test_initial(self):
        self.assertEqual(self.interpreter.configuration, self.common_states + ['i1', 'i2', 'i3', 'i4'])

    def test_parallel_order(self):
        self.interpreter.queue(Event('next'))
        step = self.interpreter.execute_once()

        self.assertEqual(self.interpreter.configuration, self.common_states + ['j1', 'j2', 'j3', 'j4'])
        self.assertEqual(step.exited_states, ['i1', 'i2', 'i3', 'i4'])
        self.assertEqual(step.entered_states, ['j1', 'j2', 'j3', 'j4'])
        self.assertEqual([t.source for t in step.transitions], ['i1', 'i2', 'i3', 'i4'])

    def test_partial_parallel_order(self):
        self.interpreter.queue(Event('next')).queue(Event('click'))
        self.interpreter.execute_once()
        step = self.interpreter.execute_once()

        self.assertEqual(self.interpreter.configuration, self.common_states + ['j1', 'j3', 'k2', 'k4'])
        self.assertEqual(step.exited_states, ['j2', 'j4'])
        self.assertEqual(step.entered_states, ['k2', 'k4'])
        self.assertEqual([t.source for t in step.transitions], ['j2', 'j4'])

    def test_partial_unnested_transition(self):
        self.interpreter.queue(Event('next')).queue(Event('reset'))
        self.interpreter.execute_once()
        step = self.interpreter.execute_once()

        self.assertEqual(self.interpreter.configuration, self.common_states + ['i1', 'i2', 'i3', 'i4'])
        self.assertLess(step.exited_states.index('r2'), step.exited_states.index('r4'))
        self.assertLess(step.exited_states.index('p1'), step.exited_states.index('p2'))
        self.assertLess(step.exited_states.index('r2'), step.exited_states.index('p1'))
        self.assertLess(step.exited_states.index('r4'), step.exited_states.index('p2'))
        self.assertLess(step.entered_states.index('p1'), step.entered_states.index('p2'))
        self.assertLess(step.entered_states.index('p1'), step.entered_states.index('r1'))
        self.assertLess(step.entered_states.index('p1'), step.entered_states.index('r2'))
        self.assertLess(step.entered_states.index('r1'), step.entered_states.index('p2'))
        self.assertLess(step.entered_states.index('r2'), step.entered_states.index('p2'))
        self.assertLess(step.entered_states.index('p2'), step.entered_states.index('r3'))
        self.assertLess(step.entered_states.index('p2'), step.entered_states.index('r4'))
        self.assertEqual([t.source for t in step.transitions], ['r2', 'r4'])

    def test_name_order(self):
        self.interpreter.queue(Event('next')).queue(Event('click')).queue(Event('next')).queue(Event('next'))
        self.interpreter.execute_once()
        self.interpreter.execute_once()
        self.interpreter.execute_once()

        self.assertEqual(self.interpreter.configuration, self.common_states + ['k1', 'k3', 'x', 'y'])

        step = self.interpreter.execute_once()
        self.assertLess(step.exited_states.index('k1'), step.exited_states.index('k3'))
        self.assertLess(step.exited_states.index('k3'), step.exited_states.index('x'))
        self.assertLess(step.exited_states.index('x'), step.exited_states.index('y'))
        self.assertEqual(self.interpreter.configuration, self.common_states + ['k1', 'x', 'y', 'z'])
        self.assertLess(step.entered_states.index('k1'), step.entered_states.index('z'))
        self.assertLess(step.entered_states.index('z'), step.entered_states.index('x'))
        self.assertLess(step.entered_states.index('x'), step.entered_states.index('y'))

        self.assertEqual([t.source for t in step.transitions], ['k1', 'k3', 'x', 'y'])

        step = self.interpreter.queue(Event('next')).execute_once()

        self.assertLess(step.exited_states.index('k1'), step.exited_states.index('x'))
        self.assertLess(step.exited_states.index('x'), step.exited_states.index('y'))
        self.assertLess(step.exited_states.index('y'), step.exited_states.index('z'))
        self.assertEqual(self.interpreter.configuration, self.common_states + ['k1', 'x', 'y', 'z'])
        self.assertLess(step.entered_states.index('k1'), step.entered_states.index('x'))
        self.assertLess(step.entered_states.index('x'), step.entered_states.index('y'))
        self.assertLess(step.entered_states.index('y'), step.entered_states.index('z'))

        self.assertEqual([t.source for t in step.transitions], ['k1', 'x', 'y', 'z'])


class BindTests(unittest.TestCase):
    def setUp(self):
        with open('tests/yaml/simple.yaml') as f:
            sc = io.import_from_yaml(f)
        self.interpreter = Interpreter(sc)
        # Stabilization
        self.interpreter.execute_once()

    def test_bind(self):
        with open('tests/yaml/simple.yaml') as f:
            other_sc = io.import_from_yaml(f)
        other_interpreter = Interpreter(other_sc)

        self.interpreter.bind(other_interpreter)
        self.assertIn(other_interpreter.queue, self.interpreter._bound)

        self.interpreter.raise_event(InternalEvent('test'))
        self.assertTrue(self.interpreter._internal_events.pop(), Event('test'))
        self.assertTrue(other_interpreter._external_events.pop(), Event('test'))

    def test_bind_callable(self):
        with open('tests/yaml/simple.yaml') as f:
            other_sc = io.import_from_yaml(f)
        other_interpreter = Interpreter(other_sc)

        self.interpreter.bind(other_interpreter.queue)
        self.assertIn(other_interpreter.queue, self.interpreter._bound)

        self.interpreter.raise_event(InternalEvent('test'))
        self.assertTrue(self.interpreter._internal_events.pop(), Event('test'))
        self.assertTrue(other_interpreter._external_events.pop(), Event('test'))


from sismic.stories import Story, Pause
from sismic.model import Event, Statechart, MacroStep
from sismic.interpreter import Interpreter
from sismic import exceptions

from typing import List, Callable

__all__ = ['ExecutionWatcher', 'teststory_from_trace']


def _teststory_from_macrostep(macrostep: MacroStep) -> Story:
    story = Story()
    story.append(Event('step started'))

    if macrostep.event:
        story.append(Event('event consumed', event=macrostep.event))

    for microstep in macrostep.steps:
        for state in microstep.exited_states:
            story.append(Event('state exited', state=state))
        if microstep.transition:
            story.append(Event('transition processed', source=microstep.transition.source,
                               target=microstep.transition.target,
                               event=macrostep.event))
        for state in microstep.entered_states:
            story.append(Event('state entered', state=state))
        for event in microstep.sent_events:
            story.append(Event('event sent', event=event))

    story.append(Event('step ended'))
    return story


class ExecutionWatcher:
    """
    This can be used to associate a property statechart with a statechart under test.
    An instance of this class is built upon an *Interpreter* instance (the tested one).

    It provides a method, namely *watch_with* which takes a property statechart
    (and a set of optional parameters that can be used to tune the interpreter that will be built upon this
    property statechart) and returns the resulting *Interpreter* instance for this tester.

    If started (using *start*), whenever something happens during the execution of the interpreter under test,
    events are automatically sent to every associated statechart properties.
    Their internal clock are synchronized, and the context of the statechart under test is
    also exposed to the property statechart, ie. if *x* is a variable in the context of a statechart under test, then
    *context.x* is dynamically exposed to every associated property statechart.

    :param tested_interpreter: Interpreter to watch
    """
    def __init__(self, tested_interpreter: Interpreter) -> None:
        self._tested = tested_interpreter
        self._testers = []  # type: List[Interpreter]
        self._failsfast = []  # type: List[bool]
        self._started = False

        self._tested_execute_once_function = tested_interpreter.execute_once

    class DynamicContext:
        def __init__(self, interpreter: Interpreter) -> None:
            self.__interpreter = interpreter

        def __getattr__(self, item):
            try:
                return self.__interpreter.context[item]
            except KeyError:
                raise AttributeError('Property statechart\'s context has no attribute {}'.format(item))

        def __copy__(self):
            return None

    def watch_with(self, property_statechart: Statechart,
                   fails_fast: bool=False,
                   interpreter_klass: Callable[..., Interpreter]=Interpreter,
                   **kwargs) -> Interpreter:
        """
        Watch the execution of the tested interpreter with given sproperty statechart.

        *interpreter_klass* is a callable that accepts a *Statechart* instance, an *initial_context* parameter and
        any additional parameters provided to this method. This callable must return an *Interpreter* instance

        :param property_statechart: a property statechart (instance of *Statechart*)
        :param fails_fast: If True (default is False), the execution of the statechart under test will raise an AssertionError
            as soon as given property statechart reaches a final state.
        :param interpreter_klass: a callable that accepts a *Statechart* instance, an *initial_context* and any
            additional (optional) parameters provided to this method.
        :return: the interpreter instance that wraps given property statechart.
        """
        # Make context available
        context = kwargs.pop('initial_context', {})
        context['context'] = ExecutionWatcher.DynamicContext(self._tested)

        tester = interpreter_klass(property_statechart, initial_context=context, **kwargs)
        self._testers.append(tester)
        self._failsfast.append(fails_fast)

        return tester

    def start(self) -> None:
        """
        Send a *started* event to the statechart properties, and starts watching the execution of
        the statechart under test.
        """
        if self._started:
            raise exceptions.ExecutionError('Cannot start an already started watcher!')

        self._started = True
        self._tested.execute_once = self.__execute_once  # type: ignore

        for tester in self._testers:
            tester.time = self._tested.time
            tester.queue(Event('execution started')).execute()

    def stop(self) -> None:
        """
        Send a *stopped* event to the statechart properties, and stops watching the execution of the
        statechart under test.
        """
        if not self._started:
            raise exceptions.ExecutionError('Cannot stop a stopped watcher!')

        for tester in self._testers:
            tester.time = self._tested.time
            tester.queue(Event('execution stopped')).execute()

        self._started = False
        self._tested.execute_once = self._tested_execute_once_function  # type: ignore

    def __execute_once(self, *args, **kwargs) -> MacroStep:
        # Execute tester
        time = self._tested.time
        step = self._tested_execute_once_function(*args, **kwargs)  # type: ignore

        story = _teststory_from_macrostep(step) if step else Story()

        # Send to testers
        for fails_fast, tester in zip(self._failsfast, self._testers):
            tester.time = time
            story.tell(tester)
            if fails_fast and tester.final:
                raise AssertionError('Property statechart "{}" is in a final state'.format(tester.statechart.name))

        return step


def teststory_from_trace(trace: List[MacroStep]) -> Story:
    """
    Return a test story based on the given *trace*, a list of macro steps.
    See documentation to see which are the events that are generated.

    Notice that this function adds a *pause* if there is any delay between pairs of consecutive steps.

    :param trace: a list of *MacroStep* instances
    :return: A story
    """
    story = Story()
    story.append(Event('execution started'))
    time = 0  # type: float

    for macrostep in trace:
        if macrostep.time > time:
            story.append(Pause(macrostep.time - time))
        story.extend(_teststory_from_macrostep(macrostep))
        time = macrostep.time

    story.append(Event('execution stopped'))
    return story
